/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmCaTrackingSetupBuilder.h
/// \brief  Tracking setup initializer in CBM (source)
/// \since  28.08.2024
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "CbmL1DetectorID.h"
#include "KfSetupBuilder.h"

#include <mutex>
#include <tuple>

namespace cbm::ca
{
  /// \class TrackingSetupBuilder
  /// \brief Encapsulation of the kf::Setup initialization routines for CBM
  class TrackingSetupBuilder {
   public:
    /// \brief Instance access
    static TrackingSetupBuilder* Instance();

    /// \brief Makes setup object
    /// \param fldMode  Field mode (kf::EFiledMode)
    template<typename T>
    cbm::algo::kf::Setup<T> MakeSetup(cbm::algo::kf::EFieldMode fldMode)
    {
      if (!fbInitialized) {
        this->Init();
      }
      return fBuilder.MakeSetup<T>(fldMode);
    }

    /// \brief  Checks, if a tracking detector is used (is in geometry and has hits)
    bool Has(ca::EDetectorID detID) const { return fvbDetUsed[detID]; }

    // Disable copy and move
    TrackingSetupBuilder(const TrackingSetupBuilder&) = delete;
    TrackingSetupBuilder(TrackingSetupBuilder&&)      = delete;
    TrackingSetupBuilder& operator=(const TrackingSetupBuilder&) = delete;
    TrackingSetupBuilder& operator=(TrackingSetupBuilder&&) = delete;

   private:
    /// \brief Default constructor
    TrackingSetupBuilder() = default;

    /// \brief Destructor
    ~TrackingSetupBuilder() = default;

    /// \brief  Check detector presence
    void CheckDetectorPresence();

    /// \brief  Initializes the instance
    void Init();

    // Material map creator properties (TODO: Provide setters, if needed)
    static constexpr double kMatCreatorPitch{0.1};    ///< Material budget map minimal bin size [cm]
    static constexpr int kMatCreatorMaxNbins{100};    ///< Max number of bins in the material budget map in x(y) axis
    static constexpr int kMatCreatorNrays{3};         ///< Number of rays per dimension for the material budget
    static constexpr bool kMatCreatorSafeMode{true};  ///< Safe mode of the material map creation
    static constexpr double kTargFieldInitStep{2.5};  ///< Step between nodes in the target field initialization [cm]
    static constexpr double kTargMaterialOffset{1};   ///< Offset between target upper limit and its material zMax [cm]

    inline static TrackingSetupBuilder* fpInstance{nullptr};
    inline static std::mutex fMutex{};

    cbm::algo::kf::SetupBuilder fBuilder{};  ///< KF-setup builder
    DetIdArr_t<bool> fvbDetUsed{{false}};    ///< Detector subsystem usage flag

    /// \brief Checks, if the setup was already initialized
    /// \note  Each call of the setup initializer resets the setup builder, so the initialization is called
    ///        in the next MakeSetup call
    bool fbInitialized{false};  ///< Check, if the setup builder initialized
  };


}  // namespace cbm::ca
