/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmGenerateMaterialMaps.h
/// \brief  A FAIR-task to generate material maps (header)
/// \since  18.02.2024
/// \author Sergei Zharko <s.zharko@gsi.de>

#include "CbmGenerateMaterialMaps.h"

#include "CbmL1.h"  // for FindTargetNode
#include "CbmTrackingDetectorInterfaceBase.h"
#include "CbmTrackingDetectorInterfaceInit.h"
#include "TFile.h"
#include "TH2F.h"

// ---------------------------------------------------------------------------------------------------------------------
//
CbmGenerateMaterialMaps::CbmGenerateMaterialMaps() : CbmGenerateMaterialMaps("CbmGenerateMaterialMaps", 0) {}

// ---------------------------------------------------------------------------------------------------------------------
//
CbmGenerateMaterialMaps::CbmGenerateMaterialMaps(const char* name, int verbose) : FairTask(name, verbose) {}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmGenerateMaterialMaps::Finish() {}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmGenerateMaterialMaps::Init()
{
  // ----- Init configuration
  if (!fsUserConfig.empty()) {
    fConfig = cbm::algo::yaml::ReadFromFile<Config>(fsUserConfig);
  }
  else {
    LOG(info) << fName << ": configuration file was not provided. Using default settings";
  }

  InitTarget();

  fpMaterialHelper = std::make_unique<cbm::ca::tools::MaterialHelper>();
  if (!fConfig.fbParallelRays) {
    fpMaterialHelper->SetDoRadialProjection(fTargetZ);
  }
  fpMaterialHelper->SetSafeMaterialInitialization(true);
  fpMaterialHelper->SetNraysPerDim(fConfig.fNofRays);

  if (fConfig.fbTrackingStations) {
    std::set<MaterialSlice> mSlice;
    // Loop over stations
    auto* pIfs = CbmTrackingDetectorInterfaceInit::Instance();
    if (pIfs) {
      for (const auto* pIf : pIfs->GetActiveInterfaces()) {
        auto detName = pIf->GetDetectorName();
        LOG(info) << fName << ": Generating material budget map for " << detName;
        int nSt = pIf->GetNtrackingStations();
        for (int iSt = 0; iSt < nSt; ++iSt) {
          MaterialSlice slice;
          slice.fName  = detName + Form("_station_%d", iSt);
          slice.fRefZ  = pIf->GetZref(iSt);
          slice.fMinZ  = pIf->GetZmin(iSt);
          slice.fMaxZ  = pIf->GetZmax(iSt);
          slice.fMaxXY = std::max(std::fabs(pIf->GetXmax(iSt)), std::fabs(pIf->GetYmax(iSt)));
          mSlice.insert(slice);
        }
      }
      double zLast = fTargetZ + 1.;
      // Run material helper
      for (auto it = mSlice.begin(); it != mSlice.end(); ++it) {
        LOG(info) << "Creating material map for " << it->fName;
        double z1   = it->fMaxZ;
        double z2   = z1;
        auto itNext = std::next(it, 1);
        if (itNext != mSlice.end()) {
          z2 = itNext->fMinZ;
        }
        double zRef  = it->fRefZ;
        double zNew  = 0.5 * (z1 + z2);
        double xyMax = kXYoffset * it->fMaxXY;
        int nBins    = static_cast<int>(std::ceil(2. * xyMax / fConfig.fPitch));
        if (nBins < 1) {
          LOG(fatal) << fName << ": selected pitch " << fConfig.fPitch << " gives wrong number of bins = " << nBins;
        }
        if (nBins > fConfig.fMaxNofBins) {
          nBins = fConfig.fMaxNofBins;
        }
        fmMaterial[it->fName] = fpMaterialHelper->GenerateMaterialMap(zRef, zLast, zNew, xyMax, nBins);
        zLast                 = zNew;
      }
    }
    else {
      LOG(error) << fName << ": tracking detector interfaces are not defined, so the materials maps cannot be "
                 << "generated for the tracking stations";
    }
  }

  // Loop over user-defined slices
  for (const auto& slice : fConfig.fvUserSlices) {
    LOG(info) << "Creating material map for " << slice.fName;
    if (fmMaterial.find(slice.fName) != fmMaterial.end()) {
      LOG(warn) << fName << ": Material for slice " << slice.fName << " was already prepared. "
                << "Please, check your configuration file";
      continue;
    }
    double xyMax = kXYoffset * slice.fMaxXY;
    double zMin  = slice.fMinZ;
    double zMax  = slice.fMaxZ;
    double zRef  = slice.fRefZ;

    int nBins = static_cast<int>(std::ceil(2. * xyMax / fConfig.fPitch));

    if (nBins < 1) {
      LOG(fatal) << fName << ": selected pitch " << fConfig.fPitch << " gives wrong number of bins = " << nBins;
    }
    if (nBins > fConfig.fMaxNofBins) {
      nBins = fConfig.fMaxNofBins;
    }
    fmMaterial[slice.fName] = fpMaterialHelper->GenerateMaterialMap(zRef, zMin, zMax, xyMax, nBins);
  }

  WriteMaterialMaps();
  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmGenerateMaterialMaps::ReInit() { return kSUCCESS; }

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmGenerateMaterialMaps::InitTarget()
{
  // Loop over all nodes till a node with name "target" is found
  // Extract the required infrmation from the node and store it in the
  // proper structure
  // The complete logic depends on the naming convention of the target.
  // If the node doesn't contain the string target the procedure will fail

  fTargetX = 1.e10;
  fTargetY = 1.e10;
  fTargetZ = 1.e10;

  TString targetPath;
  TGeoNode* targetNode{nullptr};
  CbmL1::FindTargetNode(targetPath, targetNode);

  if (!targetNode) {
    LOG(fatal) << fName << ": target node is not found";
  }

  Double_t local[3] = {0., 0., 0.};  // target centre, local c.s.
  Double_t global[3];                // target centre, global c.s.
  gGeoManager->cd(targetPath);
  gGeoManager->GetCurrentMatrix()->LocalToMaster(local, global);
  fTargetX = global[0];
  fTargetY = global[1];
  fTargetZ = global[2];

  LOG(info) << fName << ": target found: \"" << targetPath << "\" at ( " << fTargetX << " " << fTargetY << " "
            << fTargetZ << " ) [cm] ";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmGenerateMaterialMaps::WriteMaterialMaps()
{
  auto f = TFile(fsOutputFile.c_str(), "RECREATE");
  f.cd();
  for (const auto& [name, material] : fmMaterial) {
    double zMin       = material.GetZmin();
    double zMax       = material.GetZmax();
    std::string title = Form("Material thickness in X_{0}, z region [%.2f,%.2f] cm (%s)", zMin, zMax, name.c_str());
    if (fConfig.fbParallelRays) {
      title += "; horizontal projection";
    }
    else {
      title += Form("; radial projection from z=%f cm", fTargetZ);
    }
    double nBins = material.GetNbins();
    double xyMax = material.GetXYmax();
    auto* h      = new TH2F(name.c_str(), title.c_str(), nBins, -xyMax, xyMax, nBins, -xyMax, xyMax);
    h->GetXaxis()->SetTitle("x [cm]");
    h->GetYaxis()->SetTitle("y [cm]");
    for (int iBinX = 0; iBinX < nBins; ++iBinX) {
      for (int iBinY = 0; iBinY < nBins; ++iBinY) {
        h->SetBinContent(iBinX + 1, iBinY + 1, material.GetRadThickBin(iBinX, iBinY));
      }
    }
  }
  f.Write();
}
